--[[
BuyableSteeringAxle

Script to buy steering axle for trailers with an configuration

@Author:	Ifko[nator]
@Date:		07.10.2022
@Version:	v3.0

@History:	v1.0 @06.12.2017 - initial release in FS 17
			----------------------------------------------------------------------------------------------------------------
			v1.5 @08.03.2018 - support for steering axles with an steering node (Thanks to Crouwler who tell me about this!)
			----------------------------------------------------------------------------------------------------------------
			v2.0 @09.02.2020 - convert to FS 19
			----------------------------------------------------------------------------------------------------------------
			v3.0 @07.10.2022 - convert to FS 22
]]

BuyableSteeringAxle = {};
BuyableSteeringAxle.currentModName = g_currentModName;

function BuyableSteeringAxle.initSpecialization()
	local schemaSavegame = Vehicle.xmlSchemaSavegame;

	schemaSavegame:register(XMLValueType.BOOL, "vehicles.vehicle(?)." .. BuyableSteeringAxle.currentModName .. ".buyableSteeringAxle#lockSteeringAxle", "Steering Axle is locked.", false);
end;

function BuyableSteeringAxle.prerequisitesPresent(specializations)
    return SpecializationUtil.hasSpecialization(Wheels, specializations);
end;

function BuyableSteeringAxle.registerEventListeners(vehicleType)
	local functionNames = {
		"onLoad",
		"saveToXMLFile",
		"onUpdate",
		"onRegisterActionEvents",
		"onWriteStream",
		"onReadStream"
	};
	
	for _, functionName in ipairs(functionNames) do
		SpecializationUtil.registerEventListener(vehicleType, functionName, BuyableSteeringAxle);
	end;
end;

function BuyableSteeringAxle.registerFunctions(vehicleType)
	local newFunctions = {
		"setSteeringAxle"
	};
	
	for _, newFunction in ipairs(newFunctions) do
		SpecializationUtil.registerFunction(vehicleType, newFunction, BuyableSteeringAxle[newFunction]);
	end;
end;

function BuyableSteeringAxle:onLoad(savegame)
	if FlieglTrailerUtil == nil then
        return;
    end;

	local specBuyableSteeringAxle = FlieglTrailerUtil.getSpecByName(self, "buyableSteeringAxle");
	local specWheels = FlieglTrailerUtil.getSpecByName(self, "wheels");
	local specAttachable = FlieglTrailerUtil.getSpecByName(self, "attachable");

	specBuyableSteeringAxle.isBought = false;
	specBuyableSteeringAxle.lockSteeringAxle = false;
	specBuyableSteeringAxle.steerAxleIsLocked = false;
	
	if specWheels == nil or #specWheels.wheels == 0 or specAttachable.steeringAxleTargetNode ~= nil or specAttachable.steeringAxleReferenceComponentNode ~= nil then
		return;
	end;
	
	specBuyableSteeringAxle.wheels = {};
    
    local steeringAxleNumber = 0;
    
    while true do
        local steeringAxleKey = "vehicle.buyableSteeringAxles.buyableSteeringAxle(" .. tostring(steeringAxleNumber) .. ")";
        
        if not self.xmlFile:hasProperty(steeringAxleKey) then
            break;
        end;
        
        local steeringWheel = {};
        
        local wheelNumber = getXMLInt(self.xmlFile.handle, steeringAxleKey .. "#wheelNumber");
		
        if wheelNumber ~= nil then
			steeringWheel.wheelNumber = wheelNumber;

			steeringWheel.configurationName = getXMLString(self.xmlFile.handle, steeringAxleKey .. "#configurationName");
			steeringWheel.activeConfiguration = getXMLString(self.xmlFile.handle, steeringAxleKey .. "#activeConfiguration");
        	
            if steeringWheel.configurationName ~= nil and steeringWheel.configurationName ~= "" and steeringWheel.activeConfiguration ~= nil and steeringWheel.activeConfiguration ~= "" then
				steeringWheel.steeringAxleScale = Utils.getNoNil(getXMLFloat(self.xmlFile.handle, steeringAxleKey .. "#steeringAxleScale"), 0.3);
                steeringWheel.steeringAxleRotMin = MathUtil.degToRad(Utils.getNoNil(getXMLFloat(self.xmlFile.handle, steeringAxleKey .. "#steeringAxleRotMin"), -15));
                steeringWheel.steeringAxleRotMax = MathUtil.degToRad(Utils.getNoNil(getXMLFloat(self.xmlFile.handle, steeringAxleKey .. "#steeringAxleRotMax"), 15));
				
				local steeringNode = I3DUtil.indexToObject(self.components, getXMLString(self.xmlFile.handle, steeringAxleKey .. "#steeringNode"), self.i3dMappings);
				
				if steeringNode ~= nil then
					steeringWheel.steeringNode = steeringNode;
					steeringWheel.steeringNodeMinTransX = Utils.getNoNil(getXMLFloat(self.xmlFile.handle, steeringAxleKey .. "#steeringNodeMinTransX"), 0.116);
					steeringWheel.steeringNodeMaxTransX = Utils.getNoNil(getXMLFloat(self.xmlFile.handle, steeringAxleKey .. "#steeringNodeMaxTransX"), -0.116);
				end;
				
                table.insert(specBuyableSteeringAxle.wheels, steeringWheel);
            end;
        end;
        
        steeringAxleNumber = steeringAxleNumber + 1;
    end;

	if #specBuyableSteeringAxle.wheels > 0 then
        local storeItem = g_storeManager:getItemByXMLFilename(self.configFileName);
        
        for _, steeringWheel in pairs(specBuyableSteeringAxle.wheels) do

			if storeItem.configurations ~= nil and storeItem.configurations[steeringWheel.configurationName] ~= nil then
				local configurations = storeItem.configurations[steeringWheel.configurationName];
                local config = configurations[self.configurations[steeringWheel.configurationName]];
                
                for wheelNumber, wheel in pairs(specWheels.wheels) do
                    if wheelNumber == steeringWheel.wheelNumber then	
						if g_i18n:hasText(steeringWheel.activeConfiguration) then
							steeringWheel.activeConfiguration = g_i18n:getText(steeringWheel.activeConfiguration);
						end;
						
						if config.name == steeringWheel.activeConfiguration then
                            wheel.steeringAxleScale = steeringWheel.steeringAxleScale;
                            wheel.steeringAxleRotMin = steeringWheel.steeringAxleRotMin;
                            wheel.steeringAxleRotMax = steeringWheel.steeringAxleRotMax;
							
							if steeringWheel.steeringNode ~= nil then
								wheel.steeringNode = steeringWheel.steeringNode;
								wheel.steeringNodeMinTransX = steeringWheel.steeringNodeMinTransX;
								wheel.steeringNodeMaxTransX = steeringWheel.steeringNodeMaxTransX;
								
								if wheel.steeringNodeMaxRot == 0 then	
									wheel.steeringNodeMaxRot = wheel.steeringAxleRotMax;
								end;
								
								if wheel.steeringNodeMinRot == -0 then
									wheel.steeringNodeMinRot = wheel.steeringAxleRotMin;
								end;
							end;

							specBuyableSteeringAxle.isBought = true;
							specAttachable.steeringAxleUpdateBackwards = true;
                        else
                            wheel.steeringAxleScale = 0;
                            wheel.steeringAxleRotMin = 0;
                            wheel.steeringAxleRotMax = 0;
							wheel.steeringNodeMaxRot = 0;
							wheel.steeringNodeMinRot = 0;
							
							if steeringWheel.steeringNode ~= nil then
								wheel.steeringNodeMinTransX = 0;
								wheel.steeringNodeMaxTransX = 0;
							end;
                        end;
                    end;
                end;
            end;
        end;
    end;

	if specBuyableSteeringAxle.isBought then
		specBuyableSteeringAxle.l10nTexts = {};

		local l10nTexts = {
			"LOCK_STEERING_AXLE",
			"UNLOCK_STEERING_AXLE"
		};
	
		for _, l10nText in pairs(l10nTexts) do
			specBuyableSteeringAxle.l10nTexts[l10nText] = g_i18n:getText(l10nText, BuyableSteeringAxle.currentModName);
		end;
	
		if savegame ~= nil and not savegame.resetVehicles then
			specBuyableSteeringAxle.lockSteeringAxle = savegame.xmlFile:getValue(savegame.key .. "." .. BuyableSteeringAxle.currentModName .. ".buyableSteeringAxle#lockSteeringAxle", specBuyableSteeringAxle.lockSteeringAxle);
		end;
	
		self:setSteeringAxle(specBuyableSteeringAxle.lockSteeringAxle, false);
	end;
end;

function BuyableSteeringAxle:onWriteStream(streamId, connection)
	if self.setSteeringAxle ~= nil then	
		local specBuyableSteeringAxle = FlieglTrailerUtil.getSpecByName(self, "buyableSteeringAxle");

		streamWriteBool(streamId, specBuyableSteeringAxle.steerAxleIsLocked);
	end;
end;

function BuyableSteeringAxle:onReadStream(streamId, connection)
	if self.setSteeringAxle ~= nil then	
		local steerAxleIsLocked = streamReadBool(streamId);

		self:setSteeringAxle(steerAxleIsLocked, true);
	end;
end;

function BuyableSteeringAxle:onRegisterActionEvents(isActiveForInput)
	local specBuyableSteeringAxle = FlieglTrailerUtil.getSpecByName(self, "buyableSteeringAxle");
	
	if self.isClient and specBuyableSteeringAxle.isBought then
		self:clearActionEventsTable(specBuyableSteeringAxle.actionEvents);
        
		if self:getIsActiveForInput(true) then
            local actionEventId;
            
			_, actionEventId = self:addActionEvent(specBuyableSteeringAxle.actionEvents, InputAction.SET_STEERING_AXLE_BUTTON, self, BuyableSteeringAxle.toggleSteeringAxleActive, false, true, false, true, nil);
			
			g_inputBinding:setActionEventTextPriority(actionEventId, GS_PRIO_NORMAL);
			g_inputBinding:setActionEventTextVisibility(actionEventId, true);
			g_inputBinding:setActionEventActive(actionEventId, false);
		end;
	end;
end;

function BuyableSteeringAxle.toggleSteeringAxleActive(self, actionName, inputValue, callbackState, isAnalog)
	local specBuyableSteeringAxle = FlieglTrailerUtil.getSpecByName(self, "buyableSteeringAxle");
	
	self:setSteeringAxle(not specBuyableSteeringAxle.lockSteeringAxle, false);
end;

function BuyableSteeringAxle:onUpdate(dt)
	local specBuyableSteeringAxle = FlieglTrailerUtil.getSpecByName(self, "buyableSteeringAxle");
	local specAttachable = FlieglTrailerUtil.getSpecByName(self, "attachable");
	
	if specBuyableSteeringAxle.isBought then
		if self.isClient then	
			specAttachable.updateSteeringAxleAngle = not specBuyableSteeringAxle.lockSteeringAxle;
			
			local lockSteeringAxleButton = specBuyableSteeringAxle.actionEvents[InputAction.SET_STEERING_AXLE_BUTTON];
			
			if lockSteeringAxleButton ~= nil and self:getIsActive() then
				local currentText = specBuyableSteeringAxle.l10nTexts.LOCK_STEERING_AXLE;

				if specBuyableSteeringAxle.lockSteeringAxle then
					currentText = specBuyableSteeringAxle.l10nTexts.UNLOCK_STEERING_AXLE;
				end;
				
				g_inputBinding:setActionEventText(lockSteeringAxleButton.actionEventId, currentText);
				g_inputBinding:setActionEventActive(lockSteeringAxleButton.actionEventId, true);
			end;
		end;

		if specBuyableSteeringAxle.lockSteeringAxle then
			if specAttachable.steeringAxleAngle ~= 0 then	
				if specAttachable.steeringAxleAngle > 0 then
					specAttachable.steeringAxleAngle = math.max(specAttachable.steeringAxleAngle - specAttachable.steeringAxleAngleSpeed / 2 * dt, 0);
				else
					specAttachable.steeringAxleAngle = math.min(specAttachable.steeringAxleAngle + specAttachable.steeringAxleAngleSpeed / 2 * dt, 0);
				end;
			end;
		end;
	end;

	if specAttachable.updateSteeringAxleAngle then
		local steeringAngle = 0;
		local baseVehicle = self:getSteeringAxleBaseVehicle();

		if (baseVehicle ~= nil or specAttachable.steeringAxleReferenceComponentNode ~= nil) and (self.movingDirection >= 0 or specAttachable.steeringAxleUpdateBackwards) then
			yRot = Utils.getYRotationBetweenNodes(self.steeringAxleNode, specAttachable.steeringAxleReferenceComponentNode or baseVehicle.steeringAxleNode);

			local scale = 1;

			if specAttachable.steeringAxleAngleScaleSpeedDependent then
				local startSpeed = specAttachable.steeringAxleAngleScaleStart;
				local endSpeed = specAttachable.steeringAxleAngleScaleEnd;

				scale = MathUtil.clamp(1 + (self:getLastSpeed() - startSpeed) * 1 / (startSpeed - endSpeed), 0, 1);
			end;

			steeringAngle = yRot * scale;
		elseif self:getLastSpeed() > 0.2 then
			steeringAngle = 0;
		end;

		if not self:getIsSteeringAxleAllowed() then
			steeringAngle = 0;
		end;

		if specAttachable.steeringAxleDistanceDelay > 0 then
			specAttachable.steeringAxleTargetAngleHistoryMoved = specAttachable.steeringAxleTargetAngleHistoryMoved + self.lastMovedDistance;

			if specAttachable.steeringAxleTargetAngleHistoryMoved > 0.1 then
				specAttachable.steeringAxleTargetAngleHistory[specAttachable.steeringAxleTargetAngleHistoryIndex] = steeringAngle;
				specAttachable.steeringAxleTargetAngleHistoryIndex = specAttachable.steeringAxleTargetAngleHistoryIndex + 1;

				if specAttachable.steeringAxleTargetAngleHistoryIndex > #specAttachable.steeringAxleTargetAngleHistory then
					specAttachable.steeringAxleTargetAngleHistoryIndex = 1;
				end;
			end;

			local lastIndex = specAttachable.steeringAxleTargetAngleHistoryIndex + 1;

			if lastIndex > #specAttachable.steeringAxleTargetAngleHistory then
				lastIndex = 1;
			end;

			specAttachable.steeringAxleTargetAngle = specAttachable.steeringAxleTargetAngleHistory[lastIndex];
		else
			specAttachable.steeringAxleTargetAngle = steeringAngle;
		end;

		local dir = MathUtil.sign(specAttachable.steeringAxleTargetAngle - specAttachable.steeringAxleAngle);
		local speed = specAttachable.steeringAxleAngleSpeed;

		if not self.finishedFirstUpdate then
			speed = 9999;
		end;

		if dir == 1 then
			specAttachable.steeringAxleAngle = math.min(specAttachable.steeringAxleAngle + dir * dt * speed, specAttachable.steeringAxleTargetAngle);
		else
			specAttachable.steeringAxleAngle = math.max(specAttachable.steeringAxleAngle + dir * dt * speed, specAttachable.steeringAxleTargetAngle);
		end;

		if specAttachable.steeringAxleTargetNode ~= nil and (self:getLastSpeed() > 0.25 or not self.finishedFirstUpdate) then
			local angle = nil;

			if specAttachable.steeringAxleTargetNodeRefAngle ~= nil then
				local alpha = MathUtil.clamp(specAttachable.steeringAxleAngle / specAttachable.steeringAxleTargetNodeRefAngle, -1, 1);

				if alpha >= 0 then
					angle = specAttachable.steeringAxleAngleMaxRot * alpha;
				else
					angle = specAttachable.steeringAxleAngleMinRot * -alpha;
				end;
			else
				angle = MathUtil.clamp(specAttachable.steeringAxleAngle, specAttachable.steeringAxleAngleMinRot, specAttachable.steeringAxleAngleMaxRot);
			end;

			setRotation(specAttachable.steeringAxleTargetNode, 0, angle * specAttachable.steeringAxleDirection, 0);

			self:setMovingToolDirty(specAttachable.steeringAxleTargetNode);
		end;
	end;
end;

function BuyableSteeringAxle:saveToXMLFile(xmlFile, key)
	local specBuyableSteeringAxle = FlieglTrailerUtil.getSpecByName(self, "buyableSteeringAxle");
	
	if specBuyableSteeringAxle.isBought then
		xmlFile:setValue(key .. "#lockSteeringAxle", specBuyableSteeringAxle.lockSteeringAxle);
	end;
end;

function BuyableSteeringAxle:setSteeringAxle(lockSteeringAxle, noEventSend)
	local specBuyableSteeringAxle = FlieglTrailerUtil.getSpecByName(self, "buyableSteeringAxle");
	
	if lockSteeringAxle ~= specBuyableSteeringAxle.lockSteeringAxle then
		specBuyableSteeringAxle.lockSteeringAxle = lockSteeringAxle;
		
		if noEventSend == nil or noEventSend == false then
			if g_server ~= nil then
				g_server:broadcastEvent(BuyableSteeringAxleEvent.new(self, lockSteeringAxle), nil, nil, self);
			else
				g_client:getServerConnection():sendEvent(BuyableSteeringAxleEvent.new(self, lockSteeringAxle));
			end;
		end;
	end;
end;

--## MP Stuff

--## turn steering axle on/off

BuyableSteeringAxleEvent = {};
BuyableSteeringAxleEvent_mt = Class(BuyableSteeringAxleEvent, Event);

InitEventClass(BuyableSteeringAxleEvent, "BuyableSteeringAxleEvent");

function BuyableSteeringAxleEvent.emptyNew()
	local self = Event.new(BuyableSteeringAxleEvent_mt);
    
	return self;
end;

function BuyableSteeringAxleEvent.new(trailer, lockSteeringAxle)
	local self = BuyableSteeringAxleEvent:emptyNew();
	
	self.trailer = trailer;
	self.steerAxleIsLocked = lockSteeringAxle;
	
	return self;
end;

function BuyableSteeringAxleEvent:readStream(streamId, connection)
	self.trailer = NetworkUtil.readNodeObject(streamId);

	self.steerAxleIsLocked = streamReadBool(streamId);

    self:run(connection);
end;

function BuyableSteeringAxleEvent:writeStream(streamId, connection)
	NetworkUtil.writeNodeObject(streamId, self.trailer);

	streamWriteBool(streamId, self.steerAxleIsLocked);
end;

function BuyableSteeringAxleEvent:run(connection)
	self.trailer:setSteeringAxle(self.steerAxleIsLocked, true);
	
	if not connection:getIsServer() then
		g_server:broadcastEvent(BuyableSteeringAxleEvent.new(self.trailer, self.steerAxleIsLocked), nil, connection, self.trailer);
	end;
end;