--[[
TensionBeltsFix

Author:		Ifko[nator]
Datum:		04.10.2022
Version:	v1.0

History:	v1.0 @04.10.2022 - initial implementation in FS 22
]]

TensionBeltsFix = {};
TensionBeltsFix.currentModDirectory = g_currentModDirectory;
TensionBeltsFix.debugPriority = 1;

TensionBeltsFix.NUM_SEND_BITS = 16;

function TensionBeltsFix.prerequisitesPresent(specializations)
    return SpecializationUtil.hasSpecialization(TensionBelts, specializations);
end;

function TensionBeltsFix.registerEventListeners(vehicleType)
    local functionNames = {
        "onPreLoad"
    };
    
    for _, functionName in ipairs(functionNames) do
        SpecializationUtil.registerEventListener(vehicleType, functionName, TensionBeltsFix);
    end;
end;

function TensionBeltsFix:onPreLoad(savegame)
    if self.spec_tensionBelts ~= nil then
        TensionBelts.onLoad = Utils.overwrittenFunction(TensionBelts.onLoad, TensionBeltsFix.onLoadTensionBeltsFix);
    end;
end;
    
function TensionBeltsFix:onLoadTensionBeltsFix(superFunc, savegame)
    local specTensionBelts = self.spec_tensionBelts;
	local tensionBeltConfigurationId = Utils.getNoNil(self.configurations.tensionBelts, 1);
	local configKey = string.format("vehicle.tensionBelts.tensionBeltsConfigurations.tensionBeltsConfiguration(%d).tensionBelts", tensionBeltConfigurationId - 1);

	ObjectChangeUtil.updateObjectChanges(self.xmlFile, "vehicle.tensionBelts.tensionBeltsConfigurations.tensionBeltsConfiguration", tensionBeltConfigurationId, self.components, self);

	specTensionBelts.hasTensionBelts = true;

	if not self.xmlFile:hasProperty(configKey) then
		specTensionBelts.hasTensionBelts = false

		local functionNamesToRemove = {
            "onPostLoad",
            "onDelete",
            "onPreDelete",
            "onReadStream",
            "onWriteStream",
            "onUpdate",
            "onUpdateTick",
            "onDraw",
            "onRegisterActionEvents"
        };
        
        for _, functionNameToRemove in ipairs(functionNamesToRemove) do
            SpecializationUtil.removeEventListener(self, functionNameToRemove, TensionBelts);
        end;

		return;
	end;

	specTensionBelts.belts = {};
	specTensionBelts.tensionBelts = {};
	specTensionBelts.singleBelts = {};
	specTensionBelts.sortedBelts = {};
	specTensionBelts.activatable = TensionBeltsActivatable.new(self);
	specTensionBelts.totalInteractionRadius = self.xmlFile:getValue(configKey .. "#totalInteractionRadius", 6);
	specTensionBelts.interactionRadius = self.xmlFile:getValue(configKey .. "#interactionRadius", 1);
	specTensionBelts.interactionBaseNode = self.xmlFile:getValue(configKey .. "#interactionBaseNode", self.rootNode, self.components, self.i3dMappings);
	specTensionBelts.activationTrigger = self.xmlFile:getValue(configKey .. "#activationTrigger", nil, self.components, self.i3dMappings);

	if specTensionBelts.activationTrigger ~= nil then
		if not CollisionFlag.getHasFlagSet(specTensionBelts.activationTrigger, CollisionFlag.TRIGGER_PLAYER) then
			Logging.xmlError(self.xmlFile, "Missing player collision mask bit '%d' for tension belt trigger.", CollisionFlag.getBit(CollisionFlag.TRIGGER_PLAYER));
		else
			addTrigger(specTensionBelts.activationTrigger, "tensionBeltActivationTriggerCallback", self);
		end;
	end;

	specTensionBelts.allowFoldingWhileFasten = self.xmlFile:getValue(configKey .. "#allowFoldingWhileFasten", true);
	specTensionBelts.isPlayerInTrigger = false;
	specTensionBelts.checkSizeOffsets = {0, 2.5, 1.5};

	specTensionBelts.numObjectsIntensionBeltRange = 0;

	local tensionBeltType = self.xmlFile:getValue(configKey .. "#tensionBeltType", "basic");
	local beltData = g_tensionBeltManager:getBeltData(tensionBeltType);

	if beltData ~= nil then
		specTensionBelts.width = self.xmlFile:getValue(configKey .. "#width");
		specTensionBelts.ratchetPosition = self.xmlFile:getValue(configKey .. "#ratchetPosition");
		specTensionBelts.useHooks = self.xmlFile:getValue(configKey .. "#useHooks", true);
		specTensionBelts.maxEdgeLength = self.xmlFile:getValue(configKey .. "#maxEdgeLength", 0.1);
		specTensionBelts.geometryBias = self.xmlFile:getValue(configKey .. "#geometryBias", 0.01);
		specTensionBelts.defaultOffsetSide = self.xmlFile:getValue(configKey .. "#defaultOffsetSide", 0.1);
		specTensionBelts.defaultOffset = self.xmlFile:getValue(configKey .. "#defaultOffset", 0);
		specTensionBelts.defaultHeight = self.xmlFile:getValue(configKey .. "#defaultHeight", 5);
		specTensionBelts.beltData = beltData;
		specTensionBelts.linkNode = self.xmlFile:getValue(configKey .. "#linkNode", nil, self.components, self.i3dMappings);
		specTensionBelts.rootNode = self.xmlFile:getValue(configKey .. "#rootNode", self.components[1].node, self.components, self.i3dMappings);
		specTensionBelts.jointNode = self.xmlFile:getValue(configKey .. "#jointNode", specTensionBelts.rootNode, self.components, self.i3dMappings);
		specTensionBelts.checkTimerDuration = 500;
		specTensionBelts.checkTimer = specTensionBelts.checkTimerDuration;

		if specTensionBelts.linkNode == nil then
			Logging.xmlError(self.xmlFile, "No tension belts link node given at %s%s", configKey, "#linkNode");

			self:setLoadingState(VehicleLoadingUtil.VEHICLE_LOAD_ERROR);

			return;
		end;

		if getRigidBodyType(specTensionBelts.jointNode) ~= RigidBodyType.DYNAMIC and getRigidBodyType(specTensionBelts.jointNode) ~= RigidBodyType.KINEMATIC then
			Logging.xmlError(self.xmlFile, "Given jointNode '" .. getName(specTensionBelts.jointNode) .. "' has invalid rigidBodyType. Have to be 'Dynamic' or 'Kinematic'! Using '" .. getName(self.components[1].node) .. "' instead!");

			specTensionBelts.jointNode = self.components[1].node;
		end;

		local rigidBodyType = getRigidBodyType(specTensionBelts.jointNode);
		
        specTensionBelts.isDynamic = rigidBodyType == RigidBodyType.DYNAMIC;
		
        local x, y, z = localToLocal(specTensionBelts.linkNode, specTensionBelts.jointNode, 0, 0, 0);
		local rx, ry, rz = localRotationToLocal(specTensionBelts.linkNode, specTensionBelts.jointNode, 0, 0, 0);

		specTensionBelts.linkNodePosition = {x, y, z};

		specTensionBelts.linkNodeRotation = {rx, ry, rz};

		specTensionBelts.jointComponent = self:getParentComponent(specTensionBelts.jointNode);

		local tensionBeltNumber = 0;

		while true do
			local tensionBeltKey = string.format(configKey .. ".tensionBelt(%d)", tensionBeltNumber);

			if not self.xmlFile:hasProperty(tensionBeltKey) then
				break;
			end;

			if #specTensionBelts.sortedBelts == 2 ^ TensionBelts.NUM_SEND_BITS then
				Logging.xmlWarning(self.xmlFile, "Max number of tension belts is " .. 2 ^ TensionBelts.NUM_SEND_BITS .. "!");

				break;
			end;

			local startNode = self.xmlFile:getValue(tensionBeltKey .. "#startNode", nil, self.components, self.i3dMappings);
			local endNode = self.xmlFile:getValue(tensionBeltKey .. "#endNode", nil, self.components, self.i3dMappings);

			if startNode ~= nil and endNode ~= nil then
				local endX, _, _ = getTranslation(endNode);

				if math.abs(endX) < 0.0001 then
					if specTensionBelts.linkNode == nil then
						specTensionBelts.linkNode = getParent(startNode);
					end;

					if specTensionBelts.startNode == nil then
						specTensionBelts.startNode = startNode;
					end;

					specTensionBelts.endNode = endNode;
					local offsetLeft = self.xmlFile:getValue(tensionBeltKey .. "#offsetLeft");
					local offsetRight = self.xmlFile:getValue(tensionBeltKey .. "#offsetRight");
					local offset = self.xmlFile:getValue(tensionBeltKey .. "#offset");
					local height = self.xmlFile:getValue(tensionBeltKey .. "#height");
					local intersectionNodes = {};
					local intersectionNumber = 0;

					while true do
						local intersectionKey = string.format(tensionBeltKey .. ".intersectionNode(%d)", intersectionNumber);

						if not self.xmlFile:hasProperty(intersectionKey) then
							break;
						end;

						local node = self.xmlFile:getValue(intersectionKey .. "#node", nil, self.components, self.i3dMappings);

						if node ~= nil then
							table.insert(intersectionNodes, node);
						end;

						intersectionNumber = intersectionNumber + 1;
					end;

					local belt = {
						id = tensionBeltNumber + 1,
						startNode = startNode,
						endNode = endNode,
						offsetLeft = offsetLeft,
						offsetRight = offsetRight,
						offset = offset,
						height = height,
						intersectionNodes = intersectionNodes
					};

					specTensionBelts.singleBelts[belt] = belt;

					table.insert(specTensionBelts.sortedBelts, belt);
				else
					Logging.xmlWarning(self.xmlFile, "x position of endNode need to be 0 for tension belt '" .. tensionBeltKey .. "'");
				end;
			end;

			tensionBeltNumber = tensionBeltNumber + 1;
		end;

		local minX = math.huge;
		local minZ = math.huge;
		local maxX = -math.huge;
		local maxZ = -math.huge;

		for _, belt in pairs(specTensionBelts.singleBelts) do
			local sx, _, sz = localToLocal(belt.startNode, specTensionBelts.interactionBaseNode, 0, 0, 0);
			local ex, _, ez = localToLocal(belt.endNode, specTensionBelts.interactionBaseNode, 0, 0, 0);
			
            minX = math.min(minX, sx, ex);
			minZ = math.min(minZ, sz, ez);
			maxX = math.max(maxX, sx, ex);
			maxZ = math.max(maxZ, sz, ez);
		end;

		specTensionBelts.interactionBasePointX = (maxX + minX) / 2;
		specTensionBelts.interactionBasePointZ = (maxZ + minZ) / 2;

		for _, belt in pairs(specTensionBelts.singleBelts) do
			local sx, _, sz = localToLocal(belt.startNode, specTensionBelts.interactionBaseNode, 0, 0, 0);
			local sl = MathUtil.vector2Length(specTensionBelts.interactionBasePointX - sx, specTensionBelts.interactionBasePointZ - sz) + 1;
			local el = MathUtil.vector2Length(specTensionBelts.interactionBasePointX - sx, specTensionBelts.interactionBasePointZ - sz) + 1;
			
            specTensionBelts.totalInteractionRadius = math.max(specTensionBelts.totalInteractionRadius, sl, el);
		end;
	else
		Logging.xmlWarning(self.xmlFile, "No belt data found for tension belt type %s", tensionBeltType);
	end;

	specTensionBelts.hasTensionBelts = #specTensionBelts.sortedBelts > 0;
	specTensionBelts.checkBoxes = {};
	specTensionBelts.objectsToJoint = {};
	specTensionBelts.isPlayerInRange = false;
	specTensionBelts.currentBelt = nil;
	specTensionBelts.areBeltsFasten = false;
	specTensionBelts.fastedAllBeltsIndex = -1;
	specTensionBelts.fastedAllBeltsState = true;

	if self.isClient then
		specTensionBelts.samples = {
			toggleBelt = g_soundManager:loadSampleFromXML(self.xmlFile, configKey .. ".sounds", "toggleBelt", self.baseDirectory, self.components, 1, AudioGroup.VEHICLE, self.i3dMappings, self),
			addBelt = g_soundManager:loadSampleFromXML(self.xmlFile, configKey .. ".sounds", "addBelt", self.baseDirectory, self.components, 1, AudioGroup.VEHICLE, self.i3dMappings, self),
			removeBelt = g_soundManager:loadSampleFromXML(self.xmlFile, configKey .. ".sounds", "removeBelt", self.baseDirectory, self.components, 1, AudioGroup.VEHICLE, self.i3dMappings, self)
		};
	end;

	specTensionBelts.texts = {warningFoldingTensionBelts = g_i18n:getText("warning_foldingNotWhileTensionBeltsFasten")};

	if not specTensionBelts.hasTensionBelts then
		local functionNamesToRemove = {
            "onPostLoad",
            "onDelete",
            "onPreDelete",
            "onReadStream",
            "onWriteStream",
            "onUpdate",
            "onUpdateTick",
            "onDraw",
            "onRegisterActionEvents"
        };
        
        for _, functionNameToRemove in ipairs(functionNamesToRemove) do
            SpecializationUtil.removeEventListener(self, functionNameToRemove, TensionBelts);
        end;

		if specTensionBelts.activationTrigger ~= nil then
			removeTrigger(specTensionBelts.activationTrigger);

			specTensionBelts.activationTrigger = nil;
		end;
	else
		g_messageCenter:subscribe(MessageType.TREE_SHAPE_CUT, self.onTensionBeltTreeShapeCut, self);
	end;
end;